/*
 * convsym.c : convert offset's symbol to EQU symbol in object file
 *             to escape linker's error
 *  Ver 0.01 2022/02/07 by skyriver
 */

#include <stdio.h>
#include <ctype.h>

#define HDMARK		0x601a
#define HDNMARK		0x601b	/* Non contiguous Segment */

#define FLAGOFST	0x8040	/* offset symbol flag */
#define FLAGEQU		0xc000	/* equ symbols flag */


#define	BUFSIZE	128
#define NAMELEN	8

FILE *fopenb();
UWORD fread();
UWORD fwrite();

/* CP/M file header */
typedef struct {
	UWORD	mark;
	LONG	TextLen;
	LONG	DataLen;
	LONG	BssLen;
	LONG	SymLen;
	LONG	reserve0;
	LONG	reserve1;
	WORD	reserve2;	/* relocation flag */
} HEADER_TYP;


/* symbol data in file */
typedef struct {
	char	name[NAMELEN];
	UWORD	flag;
	LONG	value;
} SYMBOL_TYP;


FILE *ifp,*ofp;
char *Ifname, *Ofname;
char Buf[ BUFSIZE ];
HEADER_TYP Header;

char *Help = "convert OFFSET's symbol to EQU's symbol in object file\n\
 usage : convsym ObjFile OutputFile\n\
\t    Ver0.01 2022/02/07 by skyriver";

char *ErrFrm = "? %s";

char *ErrMsg[] = {
	 ""
	,""
	,"read header err"		/* 2 */
	,"file type is not object type"	/* 3 */
	,"write header err"		/* 4 */
	,"read err"				/* 5 */
	,"write err"			/* 6 */
};



void ErrExit( pnt )
char *pnt;
{
	fprintf( stderr, ErrFrm, pnt );
	fclose( ifp );
	fclose( ofp );
	exit( 1 );
}


void DispSym( npnt )
char *npnt;
{
	char c;
	int i;

	for( i = NAMELEN; i; i-- ) {
		c = *npnt++;
		if( c == '\0' ) {
			c = ' ';
		}
		putc( c, stderr );
	}
}


int Conv()
{
	int retval, cnt;
	LONG codelen, symlen;
	UWORD size;
	SYMBOL_TYP symbol;

	retval = 0;
	if( fread( &Header, 1, sizeof( Header ), ifp ) != sizeof( Header ) ) {
		retval = 2;
	} else {
		fprintf( stderr, "convert %s to %s", Ifname, Ofname );
		fprintf( stderr, "\n  Text:%08lx  Data:%08lx  Bss:%08lx  Symbol:%08lx\n", Header.TextLen, Header.DataLen, Header.BssLen, Header.SymLen );
		codelen = Header.TextLen + Header.DataLen;
		symlen = Header.SymLen;
		if( (Header.mark != HDMARK) && (Header.mark != HDNMARK) ) {
			retval = 3;
		} else {
			if( fwrite( &Header, 1, sizeof( Header ), ofp ) != sizeof( Header ) ) {
				retval = 4;
			} else {
				if( Header.mark == HDNMARK ) {	/* if Noncontiguous Segment */
					fread( Buf, 1,  sizeof( LONG ) * 2, ifp );
					fwrite( Buf, 1, sizeof( LONG ) * 2, ofp );
				}
				while( codelen > 0 ) {
					size = codelen > BUFSIZE ? BUFSIZE : codelen;
					if( fread( Buf, 1, size, ifp ) != size ) {
						return(5);
					}
					if( fwrite( Buf, 1, size, ofp ) != size ) {
						return(6);
					}
					codelen -= size;
				}
				size = sizeof( symbol );
				cnt = 0;
				while( symlen > 0 ) {
					if( fread( &symbol, 1, size, ifp ) != size ) {
						return(5);
					}
					if( symbol.flag == FLAGOFST ) {
						symbol.flag = FLAGEQU;
						putc( ' ',stderr );
						DispSym( symbol.name );
						fprintf( stderr, " %08lx ", symbol.value );
						if( ++cnt == 4 ) {
							cnt = 0;
							putc( '\n', stderr );
						}
					}
					if( fwrite( &symbol, 1, size, ofp ) != size ) {
						return(6);
					}
					symlen -= size;
				}
				while( (size = fread( Buf, 1, BUFSIZE, ifp )) > 0 ) {
					fwrite( Buf, 1, size, ofp );
				}
			}
		}
	}
	return( retval );
}


int main( argc, argv )
int argc;
char *argv[];
{
	char *pnt;
	int retval;

	if( argc != 3 ) {	/* args isn't two */
		fprintf( stderr, Help );
		exit( 1 );
	}
	Ifname = argv[1];
	Ofname = argv[2];
	ErrFrm = "cann't open %s";
	if( (ifp = fopenb( Ifname, "r" ) ) == NULL ) {
		ErrExit( Ifname );
	}
	else if( (ofp = fopenb( Ofname, "w") ) == NULL ) {
		ErrExit( Ofname );
	} else {
		retval = Conv();
		fclose( ifp );
		fclose( ofp );
		if( retval != 0 ) {
			fprintf( stderr, "\nerror : %s", ErrMsg[retval] );
		} else {
			fprintf( stderr, "\n  %s is created.", Ofname );
		}
	}
	return( retval );
}
